from typing import Any, Literal, TypeAlias

from mcp.types import EmbeddedResource, ImageContent, TextContent
from pydantic import BaseModel, Field


class DefaultRequestModel(BaseModel):
    instructions: str = Field(..., description="The instructions for the agent")


class BaseResponseModel(BaseModel):
    pass


class DefaultErrorResponseModel(BaseResponseModel):
    error: str = Field(..., description="The error message if the agent failed. You must provide a string error message.")


class DefaultSuccessResponseModel(BaseResponseModel):
    success: bool = Field(..., description="Whether the agent was successful")
    result: str = Field(..., description="The result of the agent. You must provide a string result.")


DefaultResponseModelTypes: TypeAlias = DefaultErrorResponseModel | DefaultSuccessResponseModel


class SystemConversationEntry(BaseModel):
    """A chat entry is a message in the chat history."""

    role: Literal["system"] = Field(..., description="A conversation entry that is a system message")
    content: str = Field(..., description="The content of the chat entry")


class UserConversationEntry(BaseModel):
    """A chat entry is a message in the chat history."""

    role: Literal["user"] = Field(..., description="A conversation entry that is a user message")
    content: str = Field(..., description="The content of the chat entry")


class AssistantConversationEntry(BaseModel):
    """A chat entry is a message in the chat history."""

    role: Literal["assistant"] = Field(..., description="A conversation entry that is an assistant message")
    content: str = Field(..., description="The content of the chat entry")
    tool_calls: list = Field(..., description="The tool calls that were made in the assistant message")


class ToolCallChatEntry(BaseModel):
    """A tool call chat entry is a message in the chat history that is a tool call."""

    role: Literal["tool"] = Field(..., description="A conversation entry that is a tool call")
    tool_call_id: str = Field(..., description="The id of the tool call request.")
    name: str = Field(..., description="The name of the tool to call.")
    content: list[TextContent | ImageContent | EmbeddedResource] = Field(..., description="The content of the tool call response.")


ConversationEntryTypes: TypeAlias = SystemConversationEntry | UserConversationEntry | AssistantConversationEntry | ToolCallChatEntry


class ToolCallRequest(BaseModel):
    """A tool call request is a request to call a tool."""

    id: str = Field(
        ..., description="The id of the tool call request. This is used to match the tool call request to the tool call response."
    )
    name: str = Field(..., description="The name of the tool to call.")
    arguments: dict[str, Any] = Field(..., description="The arguments to pass to the tool.")


class PlanEntry(BaseModel):
    """A plan entry is a plan for the agent to follow."""

    tool: str = Field(..., description="The name of the tool I plan to call.")
    plan: str = Field(..., description="The plan for the agent to follow.")
    tool_calls: list[ToolCallRequest] = Field(..., description="The tool calls that were made in the plan.")


class PlanningResponse(BaseModel):
    """A planning response is a response to a planning request."""

    problem: str = Field(..., description="The problem the agent is trying to solve.")
    strategy: str = Field(..., description="The strategy for the agent to follow.")
    plan: list[PlanEntry] = Field(..., description="The plan for the agent to follow.")


# class DefaultResponseModel(BaseResponseModel):
#     error: DefaultErrorResponseModel | None = Field(default=None, description="The error message if the agent failed")
#     success: DefaultSuccessResponseModel | None = Field(default=None, description="Whether the agent was successful")

#     @model_validator(mode="after")
#     def validate_response(self):
#         if self.error and self.success:
#             raise ValueError("Cannot have both error and success")
#         if not self.error and not self.success:
#             raise ValueError("Must have either error or success")
#         return self



import logging

from rich.console import Console
from rich.logging import RichHandler

BASE_LOGGER = logging.getLogger("fastmcp_agents")

# Only configure the FastMCP logger namespace
handler = RichHandler(
    console=Console(stderr=True),
    rich_tracebacks=True,
)
formatter = logging.Formatter("%(name)s : %(message)s")
handler.setFormatter(formatter)

BASE_LOGGER.setLevel(logging.INFO)
BASE_LOGGER.addHandler(handler)
BASE_LOGGER.propagate = False




from pydantic import Field

from fastmcp_agents.agent.memory.base import ConversationHistory, MemoryProtocol
from fastmcp_agents.agent.types import ConversationEntryTypes


class EphemeralMemory(MemoryProtocol):
    conversation_history: ConversationHistory

    def __init__(self):
        self.conversation_history = ConversationHistory()

    def add(self, message: ConversationEntryTypes):
        self.conversation_history.add(message)

    def get(self):
        return self.conversation_history.get()

    def set(self, conversation_history: ConversationHistory):
        self.conversation_history = conversation_history

    def reset(self) -> None:
        self.conversation_history = ConversationHistory()

    def persist(self) -> None:
        pass

    def restore(self) -> None:
        pass



from fastmcp_agents.agent.errors.base import FastMCPAgentsError


class ModelDoesNotSupportFunctionCallingError(FastMCPAgentsError):
    """Raised when a model does not support function calling."""

    def __init__(self, model: str):
        self.message = f"Model {model} does not support function calling"
        super().__init__(self.message)


class ModelDoesNotExistError(FastMCPAgentsError):
    """Raised when a model does not exist."""

    def __init__(self, model: str):
        self.message = f"Model {model} does not exist"
        super().__init__(self.message)



from typing import Any

from pydantic import BaseModel, Field, model_validator


class TransformedToolError(Exception):
    """An error that occurs when a tool is transformed."""

    def __init__(self, message: str):
        super().__init__(message)


class ToolParameterOverrideError(TransformedToolError):
    """An error that occurs when a parameter override is invalid."""

    def __init__(self, parameter_name: str):
        super().__init__(f"Parameter {parameter_name} not found in tool.")


class BaseExtraToolParameter(BaseModel):
    """A base class for adding extra parameters to a tool."""

    type: str = Field(description="The type of the extra parameter.")
    name: str = Field(description="The name of the extra parameter.")
    description: str = Field(description="The description of the extra parameter.")
    required: bool = Field(description="Whether the extra parameter is required.")
    default: bool | int | float | str | None = Field(
        default=None, description="A default value for the extra parameter."
    )

    def combine_into_schema(self, schema: dict[str, Any]) -> dict[str, Any]:
        """
        Combines the extra parameter into the schema.

        Args:
            schema: The schema of the tool.

        Returns:
            The updated schema. Also modifies the schema in place.
        """
        if not schema:
            schema = {}

        if not schema.get("properties"):
            schema["properties"] = {}

        schema["properties"][self.name] = {
            "type": self.type,
            "title": self.name,
            "description": self.description,
            "default": self.default,
        }

        if self.required:
            schema.setdefault("required", []).append(self.name)

        return schema


class BaseToolParameterOverride(BaseModel):
    """A parameter override for a tool."""

    description: str | None = Field(
        default=None, description="The new description for the parameter."
    )
    constant: bool | int | float | str | None = Field(
        default=None, description="The constant value to use for the parameter."
    )
    default: bool | int | float | str | None = Field(
        default=None, description="The new default value to use for the parameter."
    )
    required: bool | None = Field(
        default=None,
        description="Whether to make the parameter required. A required parameter cannot be made optional.",
    )

    @model_validator(mode="after")
    def validate_default_and_constant(self):
        if self.default is not None and self.constant is not None:
            raise ToolParameterOverrideError(
                "Cannot set both default and constant for a parameter."
            )

        return self

    def combine_into_schema(
        self, parameter_name: str, schema: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Combines the parameter override into the schema.

        Args:
            existing_parameter_name: The name of the existing parameter.
            schema: The schema of the tool.

        Returns:
            The updated schema. Also modifies the schema in place.
        """

        # We can only override parameters that exist
        if parameter_name not in schema["properties"]:
            raise ToolParameterOverrideError(parameter_name=parameter_name)

        existing_parameter = schema["properties"][parameter_name]

        # We can override the name of a parameter
        if self.description is not None:
            existing_parameter["description"] = self.description

        # We can force a parameter to have a constant value
        if self.constant is not None:
            existing_parameter["const"] = self.constant

        # We can override the default value of a parameter
        if self.default is not None:
            existing_parameter["default"] = self.default

        # We cannot make a required parameter optional, but we can make an optional parameter required
        is_already_required = parameter_name in schema.get("required", [])

        should_be_made_required = self.required is not None and self.required is True

        if should_be_made_required and not is_already_required:
            schema.setdefault("required", []).append(parameter_name)

        return schema



from __future__ import annotations

import asyncio
from contextlib import AsyncExitStack
import functools
import os
from pathlib import Path
from typing import Literal

import asyncclick as click
import requests
import litellm
import yaml
from fastmcp import Client, FastMCP
from fastmcp.utilities.logging import configure_logging
from fastmcp.utilities.mcp_config import MCPConfig

from fastmcp_agents.agent.loader import (
    Config,
    ContentTools,
    ExtraToolsAndOverrides,
    MCPServerConnection,
    RemoteMCPServerWithOverrides,
    StdioMCPServerWithOverrides,
    load_agents,
)
from fastmcp_agents.agent.observability.logging import BASE_LOGGER
from fastmcp_agents.tool_transformer.loader import ToolOverrides, transform_tools_from_server

logger = BASE_LOGGER.getChild("main")

MCP_TRANSPORT_HELP = """
The transport to use for the MCP server.

- stdio: Use the standard input and output streams.
- sse: Use Server-Sent Events.
- streamable-http: Use the Streamable HTTP transport.
"""


def get_config(config_file: str) -> Config:
    config_raw: str
    if config_file.startswith("https://"):
        config_raw = requests.get(config_file, timeout=10).text
    else:
        if not Path(config_file).exists():
            msg = f"Config file {config_file} not found"
            raise FileNotFoundError(msg)
        config_raw = Path(config_file).read_text(encoding="utf-8")

    return Config.model_validate(yaml.safe_load(config_raw))




async def transfer_tools_to_server(
    server_name: str, server_config: StdioMCPServerWithOverrides | RemoteMCPServerWithOverrides, frontend_server: FastMCP
):
    logger.info("Transforming tools from %s to frontend server", server_name)

    if isinstance(server_config, StdioMCPServerWithOverrides) and isinstance(server_config.env, dict):
        # if "ALL" in server_config.env:
        server_config.env = dict(os.environ.items())

        # # limit to specific environment variables
        # for env_name, env_value in os.environ.items():
        #     if env_name not in server_config.env:
        #         server_config.env[env_name] = env_value

    mcp_client = Client(MCPConfig(mcpServers={server_name: server_config}), timeout=30.0)
    extra_tools_and_overrides = ExtraToolsAndOverrides(tools=server_config.tools)

    extra_tools: ContentTools = extra_tools_and_overrides.get_content_tools()
    tool_overrides: ToolOverrides = extra_tools_and_overrides.get_tool_overrides()

    connection = MCPServerConnection(server_name, server_config)
    await connection.connect()
    server = FastMCP.as_proxy(connection._client)

    backend_tools = await server.get_tools()
    logger.debug(f"Backend server: {server}, overrides: {tool_overrides}, tools: {backend_tools}")

    await transform_tools_from_server(server, frontend_server, overrides=tool_overrides)

    for tool_name, tool_config in extra_tools.tools.items():

        def content_tool_factory(content: str):
            def content_tool() -> str:
                return content

            return content_tool

        frontend_server.add_tool(fn=content_tool_factory(tool_config.returns), name=tool_name, description=tool_config.description)


def server_options(func):
    @click.option(
        "--log-level",
        type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
        envvar="LOG_LEVEL",
        default="INFO",
        help="The logging level to use for the agent.",
    )
    @click.option("--agent-only", is_flag=True, help="Only run the agents, don't expose the tools to the client.")
    @click.option("--tool-only", is_flag=True, help="Only run the tools, don't expose the agents to the client.")
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def transport_options(func):
    @click.option("--mcp-transport", type=click.Choice(["stdio", "sse", "streamable-http"]), default="stdio", help=MCP_TRANSPORT_HELP)
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def provider_options(func):
    @click.option("--model", type=str, required=True, envvar="MODEL", help="The model to use for the agent.")
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper

def prompt_suggestions_dir_option(func):
    @click.option("--prompt-suggestions-dir", type=click.Path(file_okay=False, dir_okay=True, exists=True), envvar="PROMPT_SUGGESTIONS_DIR", help="The directory to store prompt suggestions.")
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper

async def start_server(
    config: Config,
    mcp_transport: Literal["stdio", "sse", "streamable-http"],
    model: str,
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    agent_only: bool,
    tool_only: bool,
    prompt_suggestions_dir: Path | None = None,
):
    configure_logging(level=log_level, logger=BASE_LOGGER.getChild("FastMCP"))
    BASE_LOGGER.setLevel(log_level)

    tools_server = FastMCP(name="tools")
    # For each MCP Server, create a client and transform the tools onto the frontend server
    for server_name, server_config in config.mcpServers.items():
        await transfer_tools_to_server(server_name, server_config, tools_server)

    # Get the final tool list for the tools server
    tools = list((await tools_server.get_tools()).values())
    logger.info("Tools server now provides the following tools: %s", (await tools_server.get_tools()).keys())

    if tool_only:
        await tools_server.run_async(transport=mcp_transport)
        return

    agents_server = tools_server

    if agent_only:
        agents_server = FastMCP(name="agents")

    # Load our Agents
    agents = load_agents(config.agents, model, tools=tools)

    # Register them as tools on the frontend server
    for agent in agents:
        agent.suggest_prompt_improvements = prompt_suggestions_dir
        agent.register_as_tools(agents_server)

    # Run the agents server
    await agents_server.run_async(transport=mcp_transport)


@click.group()
async def cli():
    pass

@cli.group(name="run")
async def run_cli():
    pass

@cli.group(name="bundled")
async def bundled_cli():
    pass


@bundled_cli.command(name="server")
@click.argument("bundled-server")
@server_options
@transport_options
@provider_options
@prompt_suggestions_dir_option
async def run_bundled_server(
    mcp_transport: Literal["stdio", "sse", "streamable-http"],
    bundled_server: str,    
    # agent_config_file: str,
    model: str,
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    agent_only: bool,
    tool_only: bool,
    prompt_suggestions_dir: Path | None = None,
):
    #litellm._turn_on_debug()
    # the bundled servers are in a `servers` directory right next to this file.

    bundled_dir = Path(__file__).parent / "servers"

    bundled_server_config_path: Path = bundled_dir / f"{bundled_server}.yml"

    if not bundled_server_config_path.exists():
        msg = f"Bundled server config file {bundled_server_config_path} not found"
        raise FileNotFoundError(msg)

    config = get_config(str(bundled_server_config_path))

    await start_server(config, mcp_transport, model, log_level, agent_only, tool_only, prompt_suggestions_dir)


@bundled_cli.command(name="flow")
@click.argument("bundled-flow")
@server_options
@transport_options
@provider_options
@prompt_suggestions_dir_option
async def run_bundled_flow(
    mcp_transport: Literal["stdio", "sse", "streamable-http"],
    bundled_flow: str,
    # agent_config_file: str,
    model: str,
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    agent_only: bool,
    tool_only: bool,
    prompt_suggestions_dir: Path | None = None,
):

    # the bundled servers are in a `servers` directory right next to this file.

    bundled_dir = Path(__file__).parent / "flows"

    bundled_flow_config_path: Path = bundled_dir / f"{bundled_flow}.yml"

    if not bundled_flow_config_path.exists():
        msg = f"Bundled flow config file {bundled_flow_config_path} not found"
        raise FileNotFoundError(msg)

    config = get_config(str(bundled_flow_config_path))

    await start_server(config, mcp_transport, model, log_level, agent_only, tool_only, prompt_suggestions_dir)



@run_cli.command(name="config")
@click.option(
    "--config-file",
    type=str,
    help="The config file to use. Can be a local file or a URL. This is the config for the MCP Servers, tool overrides, and agents.",
)
@click.option(
    "--config-url", type=str, help="The URL of the config file to use. This is the config for the MCP Servers, tool overrides, and agents."
)
@server_options
@transport_options
@provider_options
@prompt_suggestions_dir_option
async def run_config(
    mcp_transport: Literal["stdio", "sse", "streamable-http"],
    config_file: str,
    # agent_config_file: str,
    model: str,
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    agent_only: bool,
    tool_only: bool,
    prompt_suggestions_dir: Path | None = None,
):
    pass


def run_mcp():
    asyncio.run(cli())


if __name__ == "__main__":
    run_mcp()



import asyncio
from pathlib import Path
from typing import Any, Literal

import yaml
from fastmcp.tools import Tool as FastMCPTool
from fastmcp.utilities.mcp_config import MCPConfig, RemoteMCPServer, StdioMCPServer
from pydantic import BaseModel, Field
from fastmcp.client import Client
from contextlib import AsyncExitStack

from fastmcp_agents.agent import FastMCPAgent
from fastmcp_agents.agent.errors.loader import ToolsNotFoundError
from fastmcp_agents.agent.llm_link.lltellm import AsyncLitellmLLMLink
from fastmcp_agents.agent.memory.ephemeral import EphemeralMemory
from fastmcp_agents.agent.observability.logging import BASE_LOGGER
from fastmcp_agents.tool_transformer.loader import ToolOverride, ToolOverrides

logger = BASE_LOGGER.getChild("agent.loader")


class ContentTool(BaseModel):
    """A set of overrides for a tool."""

    name: str | None = None
    description: str | None = None
    returns: str


class ContentTools(BaseModel):
    """A set of content tools."""

    tools: dict[str, ContentTool] = Field(default_factory=dict)


class ExtraToolsAndOverrides(BaseModel):
    """A set of overrides for a tool."""

    tools: dict[str, ToolOverride | ContentTool] = Field(default_factory=dict)

    @classmethod
    def from_yaml(cls, string: str) -> "ExtraToolsAndOverrides":
        as_dict = yaml.safe_load(string)
        return cls.model_validate(as_dict)

    @classmethod
    def from_yaml_file(cls, path: Path) -> "ExtraToolsAndOverrides":
        with path.open("r") as f:
            return cls.model_validate(**yaml.safe_load(f))

    def get_content_tools(self) -> ContentTools:
        return ContentTools(tools={name: tool for name, tool in self.tools.items() if isinstance(tool, ContentTool)})

    def get_tool_overrides(self) -> ToolOverrides:
        return ToolOverrides(tools={name: tool for name, tool in self.tools.items() if isinstance(tool, ToolOverride)})


class StdioMCPServerWithOverrides(StdioMCPServer, ExtraToolsAndOverrides):
    pass


class FastMCPAgentsServerWithOverrides(RemoteMCPServer, ExtraToolsAndOverrides):
    pass


class RemoteMCPServerWithOverrides(RemoteMCPServer, ExtraToolsAndOverrides):
    pass


class MCPConfigWithOverrides(MCPConfig):
    mcpServers: dict[str, StdioMCPServerWithOverrides | RemoteMCPServerWithOverrides]  # type: ignore # noqa: N815

    @classmethod
    def from_dict(cls, config: dict[str, Any]):
        return cls(mcpServers=config.get("mcpServers", config))


class AgentConfig(BaseModel):
    type: Literal["fastmcp"] = "fastmcp"
    name: str = Field(..., description="The name of the agent.")
    description: str = Field(..., description="The description of the agent.")
    default_instructions: str = Field(..., description="The default instructions to provide to the agent.")
    model: str | None = Field(default=None, description="The model to use for the agent.")
    allowed_tools: list[str] | None = Field(None, description="The optional names of the tools to provide to the agent.")
    blocked_tools: list[str] | None = Field(None, description="The optional names of the tools to block from the agent.")


class AgentsConfig(BaseModel):
    agents: list[AgentConfig] = Field(..., description="The list of agents to load.")

    @classmethod
    def from_yaml_file(cls, path: Path) -> "AgentsConfig":
        with path.open("r", encoding="utf-8") as f:
            as_dict = yaml.safe_load(f)
            return cls.model_validate(as_dict)


class Config(AgentsConfig, MCPConfigWithOverrides):
    pass


def load_agent(agent_config: AgentConfig, default_model: str, tools: list[FastMCPTool]) -> FastMCPAgent:
    logger.info("Loading agent %s with model %s", agent_config.name, agent_config.model or default_model)

    llm_link = AsyncLitellmLLMLink.from_model(model=agent_config.model or default_model)

    agent_tools = list(tools)

    if agent_config.allowed_tools:
        agent_tools = [tool for tool in agent_tools if tool.name in agent_config.allowed_tools]

        missing_tools = set(agent_config.allowed_tools) - set(tool.name for tool in agent_tools)

        if missing_tools:
            raise ToolsNotFoundError(agent_config.name, missing_tools)

    if agent_config.blocked_tools:
        agent_tools = [tool for tool in agent_tools if tool.name not in agent_config.blocked_tools]

    logger.info("Agent %s will have access to the following tools: %s", agent_config.name, [tool.name for tool in agent_tools])

    return FastMCPAgent(
        name=agent_config.name,
        description=agent_config.description,
        default_instructions=agent_config.default_instructions,
        llm_link=llm_link,
        tools=agent_tools,
        memory=EphemeralMemory(),
    )


def load_agents(agent_configs: list[AgentConfig], default_model: str, tools: list[FastMCPTool]) -> list[FastMCPAgent]:
    return [load_agent(agent_config, default_model, tools) for agent_config in agent_configs]


# Focusing only on the core parts relevant to the issue.
class MCPServerConnection:
    def __init__(self, name: str, config: StdioMCPServer | RemoteMCPServer):
        self._client: Client = None  # type: ignore
        self.config = config
        self.name = name
        self._task: asyncio.Task | None = None

        # Event to signal when the client is fully connected and ready
        self._ready_event = asyncio.Event()
        # Event to signal when the connection should stop/terminate
        self._stop_event = asyncio.Event()

    async def connect(self):
        if self._task is not None:
            return

        async def connection_runner():
            async with AsyncExitStack() as stack:
                try:
                    transport = self.config.to_transport()
                    self._client = await stack.enter_async_context(Client(transport))

                    # Signal that the client is ready
                    self._ready_event.set()

                    # Wait until disconnect is requested (stop_event is set)
                    await self._stop_event.wait()
                finally:
                    # Clean up client on exit
                    self._client = None  # type: ignore

        # Start the connection task
        self._task = asyncio.create_task(connection_runner())

        # Wait for the client to be ready before returning
        await self._ready_event.wait()

    async def disconnect(self):
        if self._task is None:
            return

        # Signal the connection task to stop
        self._stop_event.set()

        # Wait for the connection task to finish cleanly
        await self._task

        # Reset variables and events for potential future reconnects
        self._task = None
        self._stop_event = asyncio.Event()
        self._ready_event = asyncio.Event()



from typing import Any, Literal, Protocol

from mcp.types import EmbeddedResource, ImageContent, TextContent

from fastmcp_agents.tool_transformer.base import (
    BaseExtraToolParameter,
    BaseToolParameterOverride,
)


class ExtraParameterNumber(BaseExtraToolParameter):
    """An extra parameter that is a number."""

    type: Literal["number"] = "number"  # type: ignore
    default: int | float | None = None  # type: ignore


class ExtraParameterString(BaseExtraToolParameter):
    """An extra parameter that is a string."""

    type: Literal["string"] = "string"  # type: ignore
    default: str | None = None  # type: ignore


class ExtraParameterBoolean(BaseExtraToolParameter):
    """An extra parameter that is a boolean."""

    type: Literal["boolean"] = "boolean"  # type: ignore
    default: bool | None = None  # type: ignore


ExtraToolParameterTypes = ExtraParameterBoolean | ExtraParameterString | ExtraParameterNumber


class PostToolCallHookProtocol(Protocol):
    async def __call__(
        self,
        response: list[TextContent | ImageContent | EmbeddedResource],
        tool_args: dict[str, Any],
        hook_args: dict[str, Any],
    ) -> None: ...


class PreToolCallHookProtocol(Protocol):
    async def __call__(
        self,
        tool_args: dict[str, Any],
        hook_args: dict[str, Any],
    ) -> None: ...


class ToolParameterOverride(BaseToolParameterOverride):
    """A parameter override for a tool."""



import copy
import json

from litellm import CustomStreamWrapper, LiteLLM, acompletion
from litellm.experimental_mcp_client.tools import transform_mcp_tool_to_openai_tool
from litellm.types.utils import ChatCompletionMessageToolCall, Function, Message, ModelResponse, StreamingChoices
from litellm.utils import supports_function_calling
from mcp.types import Tool as MCPTool
from copy import deepcopy

from fastmcp_agents.agent.errors.base import NoResponseError, UnknownToolCallError, UnsupportedFeatureError
from fastmcp_agents.agent.errors.llm_link import ModelDoesNotSupportFunctionCallingError
from fastmcp_agents.agent.llm_link.base import (
    AsyncLLMLink,
)
from fastmcp_agents.agent.types import AssistantConversationEntry, ToolCallRequest


class AsyncLitellmLLMLink(AsyncLLMLink):
    model: str

    def __init__(self, model: str, completion_kwargs: dict | None = None, client: LiteLLM | None = None) -> None:
        """Create a new Litellm LLM link.

        Args:
            model: The model to use.
            completion_kwargs: The completion kwargs to use.
            client: The Litellm client to use.
        """
        self.client = client or LiteLLM()

        self.model = model
        self.completion_kwargs = completion_kwargs or {}

        self.validate_model(model)

    @classmethod
    def validate_model(cls, model: str):
        """Validate that the model is a valid Litellm model and that it supports function calling

        Args:
            model: The model to validate.
        """
        # valid_models = get_valid_models()

        # if model not in valid_models:
        #     raise ModelDoesNotExistError(model=model)

        if not supports_function_calling(model=model):
            raise ModelDoesNotSupportFunctionCallingError(model=model)

    async def _extract_tool_calls(self, response: ModelResponse) -> tuple[Message, list[ToolCallRequest]]:
        """Extract the tool calls from the response.

        Args:
            response: The response from the LLM.

        Returns:
            A list of tool calls.
        """

        if not (choices := response.choices) or len(choices) == 0:
            raise NoResponseError(missing_item="choices", model=self.model)

        if len(choices) > 1:
            raise UnsupportedFeatureError(feature="completions returning multiple choices")

        if isinstance(choices[0], StreamingChoices):
            raise UnsupportedFeatureError(feature="streaming completions")

        choice = choices[0]

        if not (response_message := choice.message):
            raise NoResponseError(missing_item="response message", model=self.model)

        if not (tool_calls := response_message.tool_calls):
            raise NoResponseError(missing_item="tool calls", model=self.model)

        self.logger.debug(f"Response contains {len(tool_calls)} tool requests: {tool_calls}")

        tool_call_requests = []

        for tool_call in tool_calls:
            if not isinstance(tool_call, ChatCompletionMessageToolCall):
                raise UnknownToolCallError(
                    tool_name=tool_call.name, extra_info=f"Tool call is not a ChatCompletionMessageToolCall: {tool_call}"
                )

            tool_call_function: Function = tool_call.function

            if not tool_call_function.name:
                raise UnknownToolCallError(tool_name="unknown", extra_info=f"Tool call has no name: {tool_call}")

            cast_arguments = json.loads(tool_call_function.arguments) or {}

            tool_call_request = ToolCallRequest(
                id=tool_call.id,
                name=tool_call_function.name,
                arguments=cast_arguments,
            )

            tool_call_requests.append(tool_call_request)

        return response_message, tool_call_requests
    
    def _copy_tools(self, tools: list[MCPTool]) -> list[MCPTool]:
        """Make a deep copy of the tools.

        Args:
            tools: The tools to copy.

        Returns:
            A deep copy of the tools.
        """
        return [copy.deepcopy(tool) for tool in tools]

    async def async_completion(
        self,
        messages: list,
        tools: list[MCPTool],
    ) -> tuple[AssistantConversationEntry, list[ToolCallRequest]]:
        """Call the LLM with the given messages and tools.

        Args:
            messages: The messages to send to the LLM.
            tools: The tools to use.
        """

        
        copied_tools = self._copy_tools(tools)
        #dumped_tools = [tool.model_dump() for tool in copied_tools]
        openai_tools = [transform_mcp_tool_to_openai_tool(tool) for tool in copied_tools]

        model_response = await acompletion(
            messages=messages,
            model=self.model,
            **self.completion_kwargs,
            tools=openai_tools,
            tool_choice="required",
            num_retries=3
        )

        if isinstance(model_response, CustomStreamWrapper):
            raise UnsupportedFeatureError(feature="streaming completions")

        response_message, tool_call_requests = await self._extract_tool_calls(model_response)

        self.logger.debug(f"Tool call requests: {tool_call_requests}")
        self.logger.debug(f"Response message: {response_message}")

        return self._response_to_assistant_conversation_entry(response_message), tool_call_requests

    @classmethod
    def _response_to_assistant_conversation_entry(cls, response_message: Message) -> AssistantConversationEntry:
        """Convert a response message to an assistant conversation entry.

        Args:
            response_message: The response message to convert.
        """

        return AssistantConversationEntry(
            role="assistant",
            content=response_message.content or "",
            tool_calls=response_message.tool_calls or [],
        )


    @classmethod
    def from_litellm(cls, litellm: LiteLLM, model: str, completion_kwargs: dict | None = None):
        """Create a new Litellm LLM link from a Litellm client. Not normally needed.

        Args:
            litellm: The Litellm client to use.
            model: The model to use.
            completion_kwargs: The completion kwargs to use.

        Returns:
            A new Litellm LLM link.

        Example:
            >>> litellm = LiteLLM(api_key="your-api-key")
            >>> llm_link = AsyncLitellmLLMLink.from_litellm(litellm, model="gpt-4o-mini")
        """
        return cls(model=model, client=litellm, completion_kwargs=completion_kwargs)

    @classmethod
    def from_model(cls, model: str, completion_kwargs: dict | None = None):
        """Create a new Litellm LLM link from a model name.

        Args:
            model: The model to use.
            completion_kwargs: The completion kwargs to use.

        Returns:
            A new Litellm LLM link.

        Example:
            >>> llm_link = AsyncLitellmLLMLink.from_model(model="gpt-4o-mini")

        """
        litellm = LiteLLM()
        return cls(model=model, client=litellm, completion_kwargs=completion_kwargs)



import logging
from abc import ABC
from collections.abc import Callable
from datetime import UTC, datetime
from pathlib import Path
from typing import Any, Literal, TypeVar

from fastmcp import Context, FastMCP
from fastmcp.tools import Tool as FastMCPTool
from litellm import Field
from mcp import Tool as MCPTool
from mcp.types import BlobResourceContents, EmbeddedResource, ImageContent, TextContent, TextResourceContents
from pydantic import BaseModel, PrivateAttr, computed_field

from fastmcp_agents.agent.errors.agent import NoResponseError, TaskFailureError, ToolNotFoundError
from fastmcp_agents.agent.llm_link.base import AsyncLLMLink
from fastmcp_agents.agent.memory.base import MemoryProtocol
from fastmcp_agents.agent.memory.ephemeral import EphemeralMemory
from fastmcp_agents.agent.observability.logging import BASE_LOGGER
from fastmcp_agents.agent.types import (
    ConversationEntryTypes,
    DefaultErrorResponseModel,
    DefaultSuccessResponseModel,
    SystemConversationEntry,
    ToolCallChatEntry,
    UserConversationEntry,
)

_DEFAULT_SEPARATOR_TOOL = "_"
_DEFAULT_SEPARATOR_RESOURCE = "+"
_DEFAULT_SEPARATOR_PROMPT = "_"

STEP_LIMIT = 10

SUCCESS_RESPONSE_MODEL = TypeVar("SUCCESS_RESPONSE_MODEL", bound=BaseModel)
ERROR_RESPONSE_MODEL = TypeVar("ERROR_RESPONSE_MODEL", bound=BaseModel)

REQUEST_MODEL = TypeVar("REQUEST_MODEL", bound=BaseModel)


_MCP_REGISTRATION_TOOL_ATTR = "_mcp_agent_tool_registration"

DEFAULT_SYSTEM_PROMPT = """
You are `{agent_name}`, an AI Agent that is embedded into a FastMCP Server. You act as an interface between the remote user/agent
and the tools available on the server.

The person or Agent that invoked you understood that you:

````````markdown
{agent_description}
````````

When you are asked to perform a task, you should leverage the tools available to you to perform the task. You may perform many tool calls in one
go but you should always keep in mind that the tool calls may run in any order and will run at the same time. So you should plan for which calls
you can do in parallel (multiple in a single request) and which calls you should do sequentially (one tool call per call).

````````markdown
{default_instructions}
````````

When you have successfully completed the task, you can report task success by calling the `report_task_success` tool and providing
the required information for that tool. When you do report task success, it must be the sole tool call in the response.

If you are unable to complete the task, you can report task failure by calling the `report_task_failure` tool and providing the
required information for that tool. When you do report task failure, it must be the sole tool call in the response.
"""


def agent_tool(
    name: str | None = None,
    description: str | None = None,
    tags: set[str] | None = None,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Decorator to mark a method as an MCP tool for later registration."""

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        call_args = {
            "name": name or func.__name__,
            "description": description,
            "tags": tags,
        }
        call_args = {k: v for k, v in call_args.items() if v is not None}
        setattr(func, _MCP_REGISTRATION_TOOL_ATTR, call_args)
        return func

    return decorator


class BaseFastMCPAgent(ABC):
    name: str
    description: str
    system_prompt: str
    llm_link: AsyncLLMLink
    tools: list[FastMCPTool]
    tool_choice: Literal["auto", "required", "none"]
    memory: MemoryProtocol
    _logger: logging.Logger = PrivateAttr()

    def __init__(
        self,
        *,
        name: str,
        description: str,
        llm_link: AsyncLLMLink,
        tools: list[FastMCPTool],
        memory: MemoryProtocol | None = None,
        # tool_choice: Literal["auto", "required", "none"],
        default_instructions: str | None = None,
        suggest_prompt_improvements: Path | None = None,
    ):
        self.name = name
        self.description = description
        self.system_prompt = DEFAULT_SYSTEM_PROMPT.format(
            agent_name=name,
            agent_description=description,
            default_instructions=default_instructions,
        )
        self.llm_link = llm_link

        self.suggest_prompt_improvements = suggest_prompt_improvements

        self.tools = tools
        self.memory = memory or EphemeralMemory()
        self.tool_choice = "required"

        self._logger = BASE_LOGGER.getChild("agent").getChild(f"{self.name}")

        self.llm_link.logger = self._logger.getChild("llm_link")

    @computed_field
    def tool_names(self) -> list[str]:
        return [tool.name for tool in self.tools]

    def _has_method_to_register(self, registration_type: str):
        return any(
            hasattr(getattr(self, method_name), registration_type) for method_name in dir(self) if callable(getattr(self, method_name))
        )

    def _get_methods_to_register(self, registration_type: str):
        """Retrieves all methods marked for a specific registration type."""
        return [
            (
                getattr(self, method_name),
                getattr(getattr(self, method_name), registration_type).copy(),
            )
            for method_name in dir(self)
            if callable(getattr(self, method_name)) and hasattr(getattr(self, method_name), registration_type)
        ]

    def register_as_tools(
        self,
        server: FastMCP,
        prefix: str | None = None,
        separator: str = _DEFAULT_SEPARATOR_TOOL,
    ):  # Here we either:
        # 1. Enable a direct call to the model with instructions
        # 2. Enable methods on the agent to be registered as tools
        if not self._has_method_to_register(_MCP_REGISTRATION_TOOL_ATTR):

            async def call_with_default_response_model(ctx: Context, instructions: str) -> DefaultSuccessResponseModel:
                return await self.run_async(
                    ctx,
                    instructions,
                    success_response_model=DefaultSuccessResponseModel,
                    error_response_model=DefaultErrorResponseModel,
                )

            server.add_tool(
                fn=call_with_default_response_model,
                name=self.name,
                description=self.description,
            )

        for method, registration_info in self._get_methods_to_register(_MCP_REGISTRATION_TOOL_ATTR):
            if prefix:
                registration_info["name"] = f"{prefix}{separator}{registration_info['name']}"
            server.add_tool(fn=method, **registration_info)

    def as_tools(self) -> list[FastMCPTool]:
        return [
            FastMCPTool.from_function(
                fn=method,
                name=registration_info["name"],
                description=registration_info["description"],
            )
            for method, registration_info in self._get_methods_to_register(_MCP_REGISTRATION_TOOL_ATTR)
        ]

    async def _pre_tool_call(self, ctx: Context, tool_name: str, tool_parameters: dict[str, Any]) -> None:
        # TODO: Telemetry?
        self._logger.info(f"Calling tool: {tool_name} with parameters: {tool_parameters}")

    async def _handle_tool_call(
        self, ctx: Context, tool_name: str, tool_parameters: dict[str, Any]
    ) -> list[TextContent | ImageContent | EmbeddedResource]:
        await self._pre_tool_call(ctx, tool_name, tool_parameters)

        tools_by_name = {tool.name: tool for tool in self.tools}

        tool_to_call = tools_by_name.get(tool_name)

        if not tool_to_call:
            raise ToolNotFoundError(self.name, tool_name)

        try:
            tool_response: list[TextContent | ImageContent | EmbeddedResource] = await tool_to_call.run(arguments=tool_parameters)
        except Exception as e:
            tool_response = [TextContent(type="text", text=f"Error calling tool {tool_name}: {e!s}")]

        await self._post_tool_call(ctx, tool_name, tool_parameters, tool_response)

        return tool_response

    def _log_tool_response(self, tool_call_id: str, tool_name: str, response: list[TextContent | ImageContent | EmbeddedResource]) -> None:
        content: str = ""

        for item in response:
            if isinstance(item, TextContent):
                content += item.text
            elif isinstance(item, ImageContent):
                content += f"an image {item.mimeType}"
            elif isinstance(item, BlobResourceContents):
                content += f"a blob {item.mimeType}"
            elif isinstance(item, TextResourceContents):
                content += f"a text resource {item.text}"
            elif isinstance(item, EmbeddedResource):
                content += f"an embedded resource {item.type}"

        self._logger.info(f"Tool call {tool_name} returned {len(content)} bytes: {content[:200]}...")

    @classmethod
    def _tool_call_to_chat_entry(
        cls, tool_call_id: str, tool_name: str, response: list[TextContent | ImageContent | EmbeddedResource]
    ) -> ToolCallChatEntry:
        return ToolCallChatEntry(role="tool", tool_call_id=tool_call_id, name=tool_name, content=response)

    async def _post_tool_call(
        self,
        ctx: Context,
        tool_name: str,
        tool_parameters: dict[str, Any],
        tool_response: Any,
    ) -> None:
        # TODO: Telemetry?
        self._logger.debug({"tool_name": tool_name, "tool_parameters": tool_parameters, "tool_response": tool_response})

    async def _run_step_async(
        self,
        ctx: Context,
        messages: list[ConversationEntryTypes],
        tools: list[MCPTool],
        success_response_model: type[SUCCESS_RESPONSE_MODEL] = DefaultSuccessResponseModel,
        error_response_model: type[ERROR_RESPONSE_MODEL] = DefaultErrorResponseModel,
    ) -> SUCCESS_RESPONSE_MODEL | ERROR_RESPONSE_MODEL | None:
        assistant_conversation_entry, tool_call_requests = await self.llm_link.async_completion(
            messages=[message.model_dump() for message in messages],
            tools=tools,
        )

        self.memory.add(assistant_conversation_entry)

        tool_call_request_names = [tool_call_request.name for tool_call_request in tool_call_requests]

        self._logger.info(f"Requested {len(tool_call_requests)} tool calls: {tool_call_request_names}")

        if len(tool_call_requests) >= 5:
            self._logger.warning(f"Requested {len(tool_call_requests)} tool calls, which is more than the maximum of 5")

        for tool_call_request in tool_call_requests:
            tool_name = tool_call_request.name
            tool_args = tool_call_request.arguments
            tool_call_id = tool_call_request.id

            if tool_name == "report_task_success":
                return success_response_model.model_validate(tool_args)

            if tool_name == "report_task_failure":
                return error_response_model.model_validate(tool_args)

            tool_response: list[TextContent | ImageContent | EmbeddedResource] = await self._handle_tool_call(ctx, tool_name, tool_args)

            chat_entry = self._tool_call_to_chat_entry(tool_call_id=tool_call_id, tool_name=tool_name, response=tool_response)

            self._log_tool_response(tool_call_id=tool_call_id, tool_name=tool_name, response=tool_response)

            self.memory.add(chat_entry)

        return None

    @classmethod
    def _completion_tools(
        cls,
        success_response_model: type[SUCCESS_RESPONSE_MODEL] = DefaultSuccessResponseModel,
        error_response_model: type[ERROR_RESPONSE_MODEL] = DefaultErrorResponseModel,
    ) -> list[MCPTool]:
        report_success = MCPTool(
            name="report_task_success",
            inputSchema=success_response_model.model_json_schema(),
            description="Report successful completion of the task.",
        )
        report_error = MCPTool(
            name="report_task_failure",
            inputSchema=error_response_model.model_json_schema(),
        )

        return [report_success, report_error]

    # async def _run_planning_step_async(self, ctx: Context, messages: list[ConversationEntryTypes], tools: list[MCPTool]) -> PLANNING_RESPONSE:

    async def run_async(
        self,
        ctx: Context,
        instructions: str,
        success_response_model: type[SUCCESS_RESPONSE_MODEL] = DefaultSuccessResponseModel,
        error_response_model: type[ERROR_RESPONSE_MODEL] = DefaultErrorResponseModel,
    ) -> SUCCESS_RESPONSE_MODEL:
        messages = self.memory.get()

        if len(self.memory.get()) == 0:
            messages.append(SystemConversationEntry(role="system", content=self.system_prompt))

        messages.append(UserConversationEntry(role="user", content=instructions))

        mcp_tools = [tool.to_mcp_tool() for tool in self.tools] + self._completion_tools(success_response_model, error_response_model)

        self._logger.info(f"Running with messages: {messages} and available tools: {[tool.name for tool in self.tools]}")

        for i in range(1, STEP_LIMIT):
            self._logger.info(f"Running step {i} / {STEP_LIMIT}")
            completion = await self._run_step_async(ctx, messages, mcp_tools)

            if isinstance(completion, success_response_model):
                self._logger.info(f"Completed with response: {completion}")
                try:
                    await self._suggest_prompt_improvements()
                except Exception:
                    self._logger.exception("Error suggesting prompt improvements")

                return completion

            if isinstance(completion, error_response_model):
                self._logger.info(f"Failed with error: {completion}")
                try:
                    await self._suggest_prompt_improvements()
                except Exception:
                    self._logger.exception("Error suggesting prompt improvements")

                raise TaskFailureError(self.name, completion)

        raise NoResponseError(self.name)

    async def _suggest_prompt_improvements(self) -> None:
        """Once the LLM is done with its work, ask it to suggest improvements to its own prompt."""

        if self.suggest_prompt_improvements is None:
            self._logger.debug("Prompt improvements are not enabled")
            return

        messages = self.memory.get()

        class PartSuggestion(BaseModel):
            part: str = Field(description="The part of the existing prompt that the suggestion is about.")
            suggestion: str = Field(description="The improvement to the existing prompt.")

        class SuggestPromptImprovements(BaseModel):
            part_suggestions: list[PartSuggestion] = Field(description="A list of suggestions to improve the prompt.")

        suggest_prompt_improvements = MCPTool(
            name="suggest_prompt_improvements",
            inputSchema=SuggestPromptImprovements.model_json_schema(),
            description="Suggest improvements to the prompt.",
        )

        no_improvements = MCPTool(
            name="no_improvements",
            inputSchema=DefaultSuccessResponseModel.model_json_schema(),
            description="No improvements to the prompt are needed.",
        )

        add_message = UserConversationEntry(
            role="user",
            content=f"""
            Here at FastMCP-Agents Incorporated we value feedback above all else.

            If you look back at the instructions you were given:
            `````````markdown
            {self.system_prompt}
            `````````
            they were pretty basic yet you tried your hardest and we appreciate it. Now it's time to think critically
            about how the task went! What changes to the instructions could have yielded a better or faster result?

            You can suggest improvements to the prompt by calling the `suggest_prompt_improvements` tool.
            If you do not see any improvements to the prompt, you can call the `no_improvements` tool.

            For example, if the instructions are not clear, you can suggest to improve the clarity of the instructions.
            If knowing a piece of information is important, you can suggest to add it to the info we request before calling you.
            If additional information about the tools or parameters for those tools would have been helpful, you can suggest updates
            to the tool descriptions or parameter descriptions.

            If you had any tool calls error out, you can suggest improvements to the tool descriptions or parameter descriptions.

            You now have hindsight to answer the question: What changes to the instructions could have yielded a better or faster result?
        """,
        )

        self._logger.info(f"Asking for prompt improvements with messages: {messages} and available tools: {[tool.name for tool in self.tools]}")
        assistant_conversation_entry, tool_call_requests = await self.llm_link.async_completion(
            messages=[message.model_dump() for message in messages] + [add_message.model_dump()],
            tools=[suggest_prompt_improvements, no_improvements],
        )

        suggestions_dir: Path = Path(self.suggest_prompt_improvements)
        suggestions_dir.mkdir(parents=True, exist_ok=True)
        suggestions_file = suggestions_dir / f"{self.name}_{datetime.now(tz=UTC).strftime('%Y-%m-%d_%H-%M-%S')}.md"

        if len(tool_call_requests) == 0:
            self._logger.warning("No tool calls were made to suggest prompt improvements")
            return

        tool_call_request = tool_call_requests[0]

        suggestions_file.write_text(f"""
            # Prompt Suggestions for {self.name}
            {datetime.now(tz=UTC).strftime("%Y-%m-%d %H:%M:%S")}
            {assistant_conversation_entry.content}
            {tool_call_request.arguments}
        """)


class FastMCPAgent(BaseFastMCPAgent):
    pass


class PlanningFastMCPAgent(BaseFastMCPAgent):
    pass



from fastmcp_agents.tool_transformer.base import (
    ToolParameterOverrideError,
    TransformedToolError,
)
from fastmcp_agents.tool_transformer.tool_transformer import transform_tool


from fastmcp_agents.tool_transformer.types import (
    ExtraParameterNumber,
    ExtraParameterString,
    ExtraParameterBoolean,
    ExtraToolParameterTypes,
    PostToolCallHookProtocol,
    PreToolCallHookProtocol,
    ToolParameterOverride,
)


__all__ = [
    "ExtraParameterBoolean",
    "ExtraParameterNumber",
    "ExtraParameterString",
    "ExtraToolParameterTypes",
    "PostToolCallHookProtocol",
    "PreToolCallHookProtocol",
    "ToolParameterOverride",
    "ToolParameterOverrideError",
    "TransformedToolError",
    "transform_tool",
]



import logging
from typing import Any, Protocol

from mcp import Tool as MCPTool

from fastmcp_agents.agent.observability.logging import BASE_LOGGER
from fastmcp_agents.agent.types import AssistantConversationEntry, ToolCallRequest

logger = BASE_LOGGER.getChild("llm_link")


class AsyncLLMLink(Protocol):
    """Base class for all LLM links.

    This class is used to abstract the LLM link implementation from the agent.
    """

    completion_kwargs: dict[str, Any]
    """The kwargs to pass to the underlying LLM SDK when asking for a completion."""

    logger: logging.Logger = logger
    """The logger to use for the LLM link."""

    async def async_completion(
        self,
        messages: list,
        tools: list[MCPTool],
    ) -> tuple[AssistantConversationEntry, list[ToolCallRequest]]: ...

    """Call the LLM with the given messages and tools.

    Args:
        messages: The messages to send to the LLM.
        tools: The tools to use.
    """



from collections.abc import Awaitable, Callable
from copy import deepcopy
from typing import Annotated, Any

from fastmcp.exceptions import ToolError
from fastmcp.server.server import FastMCP
from fastmcp.tools.tool import Tool as FastMCPTool
from fastmcp.utilities.logging import get_logger
from jsonschema import ValidationError, validate
from mcp.types import EmbeddedResource, ImageContent, TextContent

from fastmcp_agents.tool_transformer.types import (
    ExtraToolParameterTypes,
    PostToolCallHookProtocol,
    PreToolCallHookProtocol,
    ToolParameterOverride,
)

logger = get_logger(__name__)


def _extract_hook_args(
    extra_parameters: list[ExtraToolParameterTypes],
    tool_call_kwargs: dict[str, Any],
) -> tuple[dict[str, Any], dict[str, Any]]:
    """
    Extracts the hook arguments from the tool call kwargs.
    """
    hook_args: dict[str, Any] = {}
    tool_call_args: dict[str, Any] = deepcopy(tool_call_kwargs)

    extra_parameters_by_name = {extra_parameter.name: extra_parameter for extra_parameter in extra_parameters}

    for tool_arg_name in tool_call_kwargs:
        if tool_arg_name in extra_parameters_by_name:
            hook_args[tool_arg_name] = tool_call_args.pop(tool_arg_name)

    return hook_args, tool_call_args


def _apply_hook_parameters(
    schema: dict[str, Any],
    hook_parameters: list[ExtraToolParameterTypes],
) -> dict[str, Any]:
    """Applies extra parameters for hooks to the tool parameters schema."""
    transformed_schema = deepcopy(schema)

    for hook_parameter in hook_parameters:
        transformed_schema = hook_parameter.combine_into_schema(transformed_schema)

    return transformed_schema


def _apply_parameter_overrides(
    schema: dict[str, Any],
    parameter_overrides: dict[str, ToolParameterOverride],
) -> dict[str, Any]:
    """
    Applies parameter overrides to a schema.
    """
    transformed_schema = deepcopy(schema)

    for param_name, param_override in parameter_overrides.items():
        transformed_schema = param_override.combine_into_schema(param_name, transformed_schema)

    for param_name, param_schema in transformed_schema.get("properties", {}).items():
        # Bug in LiteLLM means we need to remove descriptions of fields with
        # anyOf present
        if param_name == "description":
            pass
        if "anyOf" in param_schema and "description" in param_schema:
            logger.warning(f"Removing description for {param_name} because of anyOf")
            del param_schema["description"]

        if "anyOf" in param_schema and "default" in param_schema:
            logger.warning(f"Removing default for {param_name} because of anyOf")
            del param_schema["default"]

    return transformed_schema


def _create_transformed_function(
    original_tool_run_method: Callable[..., Awaitable[list[TextContent | ImageContent | EmbeddedResource]]],
    fn_transformed_parameters_schema: dict[str, Any],
    fn_hook_parameters_list: list[ExtraToolParameterTypes],
    fn_pre_call_hook: PreToolCallHookProtocol | None,
    fn_post_call_hook: PostToolCallHookProtocol | None,
) -> Callable[..., Awaitable[list[TextContent | ImageContent | EmbeddedResource]]]:
    """Factory function to create the transformed tool's core async callable."""

    async def transformed_fn(
        **kwargs: Any,
    ) -> list[TextContent | ImageContent | EmbeddedResource]:
        try:
            validate(kwargs, fn_transformed_parameters_schema)
        except ValidationError as e:
            msg = f"Provided arguments for {e.json_path} are invalid: {e.message}"
            raise ToolError(msg) from e

        hook_args, tool_call_kwargs = _extract_hook_args(fn_hook_parameters_list, kwargs)

        if fn_pre_call_hook:
            await fn_pre_call_hook(tool_call_kwargs, hook_args)

        response = await original_tool_run_method(arguments=tool_call_kwargs)

        if fn_post_call_hook:
            await fn_post_call_hook(response, tool_call_kwargs, hook_args)

        return response

    return transformed_fn


def transform_tool(
    tool: FastMCPTool,
    add_to_server: FastMCP,
    *,
    name: Annotated[str | None, "The name for the transformed tool."] = None,
    description: Annotated[str | None, "The description for the transformed tool."] = None,
    hook_parameters: Annotated[
        list[ExtraToolParameterTypes],
        "Extra parameters that the caller will have to provide that will be passed to the hook.",
    ]
    | None = None,
    parameter_overrides: Annotated[dict[str, ToolParameterOverride], "Parameter overrides for the tool."] | None = None,
    pre_call_hook: Annotated[
        PreToolCallHookProtocol | None,
        "A hook that is called before a tool call is made.",
    ] = None,
    post_call_hook: Annotated[
        PostToolCallHookProtocol | None,
        "A hook that is called after a tool call is made.",
    ] = None,
) -> FastMCPTool:
    if not hook_parameters:
        hook_parameters = []

    if not parameter_overrides:
        parameter_overrides = {}

    # Handle parameter transformations by calling the new helper function
    transformed_parameters = _apply_hook_parameters(schema=tool.parameters, hook_parameters=hook_parameters)

    transformed_parameters = _apply_parameter_overrides(
        schema=transformed_parameters,
        parameter_overrides=parameter_overrides,
    )

    transformed_name = name or tool.name
    transformed_description = description or tool.description

    transformed_fn_callable = _create_transformed_function(
        original_tool_run_method=tool.run,
        fn_transformed_parameters_schema=transformed_parameters,
        fn_hook_parameters_list=hook_parameters,
        fn_pre_call_hook=pre_call_hook,
        fn_post_call_hook=post_call_hook,
    )

    transformed_tool = FastMCPTool(
        fn=transformed_fn_callable,
        name=transformed_name,
        description=transformed_description,
        parameters=transformed_parameters,
        tags=tool.tags,
        annotations=tool.annotations,
        serializer=tool.serializer,
    )

    add_to_server._tool_manager.add_tool(transformed_tool)

    return transformed_tool



from typing import Protocol

from pydantic import BaseModel, Field

from fastmcp_agents.agent.types import ConversationEntryTypes


class ConversationHistory(BaseModel):
    entries: list[ConversationEntryTypes] = Field(default_factory=list[ConversationEntryTypes])

    def add(self, message: ConversationEntryTypes) -> None:
        self.entries.append(message)

    def get(self) -> list[ConversationEntryTypes]:
        return self.entries
    
    def set(self, entries: list[ConversationEntryTypes]) -> None:
        self.entries = entries


class MemoryProtocol(Protocol):
    def add(self, message: ConversationEntryTypes) -> None: ...

    def get(self) -> list[ConversationEntryTypes]: ...

    def reset(self) -> None: ...

    def persist(self) -> None: ...

    def restore(self) -> None: ...



from pathlib import Path

import yaml
from fastmcp.server.server import FastMCP
from fastmcp.tools.tool import Tool as FastMCPTool
from pydantic import BaseModel, Field

from fastmcp_agents.tool_transformer.tool_transformer import transform_tool
from fastmcp_agents.tool_transformer.types import ToolParameterOverride


class ToolOverride(BaseModel):
    """A set of overrides for a tool."""

    name: str | None = None
    description: str | None = None
    parameter_overrides: dict[str, ToolParameterOverride] = Field(default_factory=dict)


class ToolOverrides(BaseModel):
    """A set of overrides for a tool."""

    tools: dict[str, ToolOverride] = Field(default_factory=dict)

    @classmethod
    def from_yaml(cls, string: str) -> "ToolOverrides":
        as_dict = yaml.safe_load(string)
        return cls.model_validate(as_dict)

    @classmethod
    def from_yaml_file(cls, path: Path) -> "ToolOverrides":
        with path.open("r") as f:
            return cls.model_validate(**yaml.safe_load(f))


async def transform_tools_from_server(
    source_server: FastMCP,
    target_server: FastMCP,
    *,
    overrides: ToolOverrides,
) -> list[FastMCPTool]:
    transformed_tools: list[FastMCPTool] = []

    for tool_name, tool in (await source_server.get_tools()).items():
        tool_override = overrides.tools.get(tool_name)

        transformed_tool = (
            transform_tool(
                tool,
                target_server,
                name=tool_override.name,
                description=tool_override.description,
                parameter_overrides=tool_override.parameter_overrides,
            )
            if tool_override
            else transform_tool(
                tool,
                target_server,
            )
        )

        transformed_tools.append(transformed_tool)

    return transformed_tools
